#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <asm/uaccess.h>

#define DEV_NAME "linux-magazin"
#define DEV_NAME_SIMU "linux-magazin.simu"

static dev_t lm_dev_number;
static struct cdev *driver_object;
struct class *lm_class;
static struct device *lm_dev, *lm_dev_simu;

static u32 data_register;
static u32 (*mapper_readl)(void *addr);

static int driver_open(struct inode *devicefile, struct file *filp);

ssize_t application_read(struct file *instance, char __user *buffer,
    size_t max_bytes_to_read, loff_t *offset)
{
    size_t to_copy, not_copied;
    u32 reg;

    reg = mapper_readl((void *)0x12345678);
    to_copy = min(sizeof(reg),max_bytes_to_read);
    not_copied = copy_to_user(buffer, &reg, to_copy);
    return to_copy-not_copied;
}

static ssize_t simulation_write(struct file * instance, const char *userbuf,
    size_t length, loff_t *o)
{
    int to_copy, not_copied;

    to_copy = min(length, sizeof(data_register));
    not_copied=copy_from_user(&data_register, userbuf, to_copy);
    printk("simulation_write: data_register=0x%x\n", data_register);
    return to_copy-not_copied;
}

static struct file_operations fops_application = {
    .open = driver_open,
    .read = application_read,
};

static struct file_operations fops_simu = {
    .write = simulation_write,
};

static int driver_open(struct inode *devicefile, struct file *filp)
{
    if (MINOR(devicefile->i_rdev) == 0)
        filp->f_op = &fops_application;
    else
        filp->f_op = &fops_simu;
    return 0;
}

static u32 simulation_readl(void *addr)
{
    return data_register;
}

static int __init mod_init(void)
{
    mapper_readl = simulation_readl; // Simulation aktivieren

    if (alloc_chrdev_region(&lm_dev_number, 0, 2, DEV_NAME) < 0)
        return -EIO;
    driver_object = cdev_alloc();
    if (driver_object==NULL)
        goto free_device_number;
    driver_object->owner = THIS_MODULE;
    driver_object->ops = &fops_application;
    if (cdev_add(driver_object, lm_dev_number, 2))
        goto free_cdev;
    lm_class = class_create(THIS_MODULE, DEV_NAME);
    if (IS_ERR(lm_class)) {
        pr_err("template: no udev support\n");
        goto free_cdev;
    }
    lm_dev = device_create(lm_class, NULL,
        lm_dev_number, NULL, "%s", DEV_NAME);
    lm_dev_simu = device_create(lm_class, NULL,
        lm_dev_number + 1, NULL, "%s", DEV_NAME_SIMU);
    return 0;
free_cdev:
    kobject_put(&driver_object->kobj);
free_device_number:
    unregister_chrdev_region(lm_dev_number, 2);
    return -EIO;
}

static void __exit mod_exit(void)
{
    device_destroy(lm_class, lm_dev_number);
    device_destroy(lm_class, lm_dev_number + 1);
    class_destroy(lm_class);
    cdev_del(driver_object);
    unregister_chrdev_region(lm_dev_number, 2);
    return;
}

module_init(mod_init);
module_exit(mod_exit);
MODULE_LICENSE("GPL");
